package nl.fairbydesign.backend.parsers;

import nl.fairbydesign.backend.ApplicationServiceInitListener;
import nl.fairbydesign.backend.WebGeneric;
import nl.fairbydesign.backend.irods.Connection;
import nl.fairbydesign.backend.irods.Data;
import nl.fairbydesign.backend.metadata.Metadata;
import nl.munlock.ontology.domain.*;
import nl.wur.ssb.RDFSimpleCon.ResultLine;
import nl.wur.ssb.RDFSimpleCon.api.Domain;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.jena.rdf.model.NodeIterator;
import org.apache.jena.rdf.model.RDFList;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.StmtIterator;
import org.apache.poi.hssf.usermodel.HSSFDateUtil;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.irods.jargon.core.connection.IRODSAccount;
import org.irods.jargon.core.exception.DataNotFoundException;
import org.irods.jargon.core.exception.DuplicateDataException;
import org.irods.jargon.core.exception.JargonException;
import org.irods.jargon.core.protovalues.UserTypeEnum;
import org.irods.jargon.core.pub.CollectionAO;
import org.irods.jargon.core.pub.UserAO;
import org.irods.jargon.core.pub.UserGroupAO;
import org.irods.jargon.core.pub.domain.User;
import org.irods.jargon.core.pub.domain.UserGroup;
import org.jermontology.ontology.JERMOntology.domain.*;

import org.purl.ppeo.PPEO.owl.domain.observation_unit;
import org.rdfhdt.hdt.exceptions.IllegalFormatException;
import org.schema.domain.Organization;
import org.schema.domain.Person;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.Normalizer;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.vaadin.flow.shared.util.SharedUtil.splitCamelCase;
import static nl.fairbydesign.backend.WebGeneric.getStringValueCell;
import static nl.fairbydesign.backend.WebGeneric.setFileFormat;
import static nl.munlock.ontology.domain.SequencingPlatform.*;

/**
 * Main class for validating the excel file and transforming it into RDF
 */
public class ExcelValidator {
    // List of constant variables for easy modifications
    public static final String FORWARD_FILENAME = "Forward filename";
    public static final String OBSERVATION_UNIT_IDENTIFIER = "Observation unit identifier";
    public static final String SAMPLE_IDENTIFIER = "Sample identifier";
    public static final String FILE_NAME = "Filename";
    public static final String ASSAY_IDENTIFIER = "Assay identifier";
    public static final String SAME_AS = "Same as";
    public static final String LIBRARY_NAME = "Library name";
    public static final String REVERSE_FILENAME = "Reverse filename";
    public static final String ASSAY_DESCRIPTION = "Assay description";
    public static final String ASSAY_TITLE = "Assay title";
    public static final String PLATFORM = "Platform";
    public static final String REFERENCE = "Reference";
    public static final String STUDY_IDENTIFIER = "Study identifier";
    public static final String PROJECT_IDENTIFIER = "Project identifier";
    public static final String PROJECT_DESCRIPTION = "Project description";
    public static final String PROJECT_TITLE = "Project title";
    public static final String STUDY_DESCRIPTION = "Study description";
    public static final String STUDY_TITLE = "Study title";
    public static final String INVESTIGATION_IDENTIFIER = "Investigation identifier";
    public static final String INVESTIGATION_TITLE = "Investigation title";
    public static final String INVESTIGATION_DESCRIPTION = "Investigation description";
    public static final String FIRSTNAME = "Firstname";
    public static final String LASTNAME = "Lastname";
    public static final String EMAIL = "Email address";
    public static final String ORCID = "ORCID";
    public static final String DEPARTMENT = "Department";
    public static final String ORGANIZATION = "Organization";

    // RDF triple store
    private Domain domain;
    // logger
    private final Logger logger = LoggerFactory.getLogger(ExcelValidator.class);
    // RDF prefix
    private String PREFIX;
    // Message to be shown to the user
    private String message = "";
    // Obtain all information from the excel sheet template?
    private Map<String, ArrayList<Metadata>> metadataSet = ApplicationServiceInitListener.getMetadata();

    // String variables

    /**
     * @param fileName of the input excel file
     * @param excelFile input stream to be processed
     * @param irodsAccount with credentials to upload after validation succeeds when logged in (not obligatory)
     * @return the RDF object
     * @throws Exception when validation fails, the message can be expanded upon failure to show a clearer message to
     * the user
     */
    public String validate(String fileName, InputStream excelFile, IRODSAccount irodsAccount) throws Exception {
        this.setMessage("");

        logger.debug("Obtaining information from reference file");

        // Ensuring metadata set is final
        metadataSet = Collections.unmodifiableMap(metadataSet);

        logger.debug("Validation starting");

        // Opening the workbook stream
        XSSFWorkbook workbook;
        try {
            workbook = new XSSFWorkbook(excelFile);
        } catch (Exception e) {
            logger.error("Not a valid excel file " + e.getMessage());
            return "Is this an excel file?";
        }

        // Creation of an RDF memory store
        domain = new Domain("");

        // Add generic prefixes to the RDF file for more compressed TURTLE format
        setPrefixes();

        /**
         * The order of creation Project > Investigation > Study > Observation Unit > Sample > Assay is flipped around due to cardinality restrictions by the API
         * An investigation needs to have 1 or more studies
         * A study needs to have 1 or more observation units
         * etc ...
         */

        // Get host from system environment or fall back to default unlock url
        if (System.getenv("irodsHost") != null) {
            String host = System.getenv("irodsHost");
            if (host.startsWith("http")) {
                PREFIX = System.getenv("irodsHost") + "/";
            } else {
                PREFIX = "http://" + System.getenv("irodsHost") + "/";
            }
        } else {
            // TODO change to more generic prefix?
            PREFIX = "http://unlock-icat.irods.surfsara.nl/";
        }

        // Base prefix added
        domain.getRDFSimpleCon().setNsPrefix("base", PREFIX);

        boolean sampleFound = false;
        // Parsing all sample and assay types with the generic assay sheet as well
        for (int i = 0; i < workbook.getNumberOfSheets(); i++) {
            String sheetName = workbook.getSheetAt(i).getSheetName();
            // Parsing sample metadata
            if (sheetName.toLowerCase().startsWith("sample")) {
                String packageName;
                if (sheetName.toLowerCase().contains(" - ")) {
                    packageName = sheetName.split(" - ", 2)[1];
                } else {
                    packageName = "core";
                }
                sampleFound = true;

                ArrayList<Metadata> metadataArrayList = new ArrayList<>();

                metadataSet.get("Sample").forEach(e -> {
                    // Add elements from package name
                    if (e.getPackage().matches(packageName)) {
                        metadataArrayList.add(e);
                    }
                    // Add core elements if package name was not core
                    if (!packageName.matches("core") && e.getPackage().matches("core")) {
                        metadataArrayList.add(e);
                    }
                });

                // Process the sample sheet
                sampleCreation(workbook.getSheet(sheetName), metadataArrayList, packageName);
            }

            if (sheetName.toLowerCase().contains("assay")) {
                String packageName;
                if (sheetName.toLowerCase().contains(" - ")) {
                    packageName = sheetName.split(" - ", 2)[1];
                } else {
                    String name = sheetName.replace("Assay", "");
                    if (name.length() > 0) {
                        packageName = name.toLowerCase();
                    } else {
                        packageName = "core";
                    }
                }

                ArrayList<Metadata> metadataArrayList = new ArrayList<>();

                metadataSet.get("Assay").forEach(e -> {
                    // Add elements from package name
                    if (e.getPackage().equalsIgnoreCase(packageName)) {
                        metadataArrayList.add(e);
                    }
                    // Add core elements if package name was not core
                    if (!packageName.equalsIgnoreCase("core") && e.getPackage().equalsIgnoreCase("core")) {
                        metadataArrayList.add(e);
                    }
                });

                // Process the sample sheet
                assayCreation(workbook.getSheet(sheetName), metadataArrayList, packageName);
                // sampleCreation(workbook.getSheet(sheetName), metadataArrayList, packageName);
            }
            // Removes additional names of the sheet for now
            // TODO check sample processing
//            String cleanSheetName = sheetName.replaceAll(".*_", "").replaceAll(".* ", "");
//            if (cleanSheetName.toLowerCase().endsWith("assay")) {
//                logger.info("Processing " + cleanSheetName);
//                // Obtain the specific assay first due to potential copy issues
//                ArrayList<Metadata> assaySet = metadataSet.get(cleanSheetName);
//                // Obtain the parent assay
//                assaySet.addAll(metadataSet.get("Assay"));
//                // Parse the assay
//                assayCreation(workbook.getSheet(sheetName), assaySet, null);
//            }
        }

        if (!sampleFound) {
            logging("The sample sheet and package cannot be found.\n" +
                    "Please make sure there is a sheet that starts with sample followed by the package name or generate a new excel sheet with the sample sheet / package in place.");
            throw new Exception();
        }

        // Parse the observation unit information
        observationUnitCreation(workbook.getSheet("ObservationUnit"), metadataSet.get("ObservationUnit"));

        // Parse the study information
        studyCreation(workbook.getSheet("Study"), metadataSet.get("Study"));

        // Parse the investigation information
        investigationCreation(workbook.getSheet("Investigation"), metadataSet.get("Investigation"));

        // Parse the project information
        projectCreation(workbook.getSheet("Project"), metadataSet.get("Project"));

        validate();

        // Setting up folders and files in irods
        if (irodsAccount != null) {
            logger.info("User logged in, setting up irods landingzone");
            String projectIdentifier = irodsSetup(fileName, workbook, new Connection());
            return projectIdentifier;
        }

        String rdfFile = fileName.replace(".xlsx", "") + ".ttl";

        domain.save(rdfFile);

        logging("Validation successful, user not logged in. Result file not uploaded to data system");

        return rdfFile;
    }

    private String irodsSetup(String fileName, Workbook workbook, Connection connection) throws Exception {
        // Acquire the first project identifier used throughout the parsing
        String projectIdentifier = getProjectIdentifier(workbook.getSheet("Project"));
        String landingZoneProjectPath = "/" + connection.getIrodsAccount().getZone() + "/landingzone/projects/PRJ_" + projectIdentifier;

        // Create raw data folder / <project> / <investigation>
        Project project = getProject(projectIdentifier);
        if (project.getAllHasPart().size() == 0) {
            throw new Exception("No investigations found");
        }

        ArrayList<String> investigationIDs = new ArrayList<>();
        ArrayList<String> studyIDs = new ArrayList<>();
        for (Investigation investigation : project.getAllHasPart()) {
            // Get investigation identifier
            investigationIDs.add(investigation.getIdentifier());
            // Get study identifier
            investigation.getAllHasPart().forEach(study -> studyIDs.add(study.getIdentifier()));
            String landingZoneInvestigationPath = landingZoneProjectPath + "/INV_" + investigation.getIdentifier();
            logger.debug("Creating investigation " + landingZoneInvestigationPath);
            // Create directories
            connection.getFileFactory().instanceIRODSFile(landingZoneProjectPath + "/data").mkdirs();
            connection.getFileFactory().instanceIRODSFile(landingZoneInvestigationPath).mkdirs();

            // Get the researchers and assign them to this landingzone
            String groupName = project.getIdentifier() + "_" + investigation.getIdentifier();

            setUsers(connection, investigation.getAllHasContributor(), groupName, landingZoneProjectPath, landingZoneInvestigationPath);
        }

        // Create data folder
        connection.getFileFactory().instanceIRODSFile(landingZoneProjectPath + "/data").mkdir();

        Collections.sort(investigationIDs);
        String investigationID = investigationIDs.get(0);
        String studyID = studyIDs.get(0);

        logging("\n\nSuccessfully parsed the excel sheet");

        // Only perform the rest when user is logged in?
        logger.debug("USERNAME BOT: " + connection.getIrodsAccount().getUserName());

        // When successful upload to a designated location in irods
        fileName = fileName.replaceAll(".xlsx$","");
        File xlsxFile = new File(fileName + ".xlsx");
        File turtleFile = new File(fileName + ".ttl");
        File remotexlsxFile = new File(fileName.replaceAll(" +", "_") + ".xlsx");
        File remoteturtleFile = new File(fileName.replaceAll(" +", "_") + ".ttl");

        // RDF object is created, now pushed to the iRODS parser
        domain.save(turtleFile.getName());

        // Write to temp file
        OutputStream outputStream = new FileOutputStream(xlsxFile.getName());
        workbook.write(outputStream);
        outputStream.close();

        // Upload to all investigations
        for (Investigation investigation : project.getAllHasPart()) {
            logging("\n\nUploading excel file to the data management system");
            File destination = new File(landingZoneProjectPath + "/INV_" + investigation.getIdentifier() + "/");
            Data.uploadIrodsFile(new Connection().getIrodsAccount(), xlsxFile, new File(destination + "/" + remotexlsxFile.getName()));
            logging("\n\nUploading database file to the data management system");
            Data.uploadIrodsFile(new Connection().getIrodsAccount(), turtleFile, new File(destination + "/" + remoteturtleFile.getName()));
            logging("\n\nFiles uploaded to: " + destination + "/");
            logging("\nExcel file: " + xlsxFile);
            logging("\nTurtle file: " + turtleFile);

            // Technicians have own rights?
            CollectionAO collectionAO = connection.getAccessObjectFactory().getCollectionAO(connection.getIrodsAccount());
            collectionAO.setAccessPermissionOwn(connection.getIrodsAccount().getZone(), landingZoneProjectPath, "technicians", true);

            // Remove all traces of Mr.Robot
            collectionAO.removeAccessPermissionForUser(connection.getIrodsAccount().getZone(), landingZoneProjectPath, "Mr.Robot", true);
        }
        return projectIdentifier;
    }

    private void setUsers(Connection connection, List<? extends Person> allResearcher, String groupName, String landingZoneProjectPath, String landingZoneInvestigationPath) throws JargonException {
        logger.info("Setting user permissions to the landingzone");

        if (allResearcher.size() == 0) {
            throw new JargonException("No contacts found");
        }

        // Create a group
        UserGroupAO userGroupAO = connection.getAccessObjectFactory().getUserGroupAO(connection.getIrodsAccount());

        UserGroup ug = userGroupAO.findByName(groupName);
        if (ug == null) {
            logger.info("Making group: " + groupName);
            ug = new UserGroup();
            ug.setUserGroupName(groupName);
            ug.setZone(connection.getIrodsAccount().getZone());
            userGroupAO.addUserGroup(ug);
            logger.info("Group made: " + ug.getUserGroupName());
        } else {
            logger.info("Group: " + groupName);
        }

        UserAO userAO = connection.getAccessObjectFactory().getUserAO(connection.getIrodsAccount());

        // Add group to investigation landingzone
        logger.info("Adding group to project as read only");
        CollectionAO collectionAO = connection.getAccessObjectFactory().getCollectionAO(connection.getIrodsAccount());
        // Set read permission for project to false?
        collectionAO.setAccessPermissionRead(connection.getIrodsAccount().getZone(), landingZoneProjectPath, groupName, false);
        // Set write permission for investigation
        collectionAO.setAccessPermissionWrite(connection.getIrodsAccount().getZone(), landingZoneInvestigationPath, groupName, true);
        // Enable inheritance for investigation folder
        collectionAO.setAccessPermissionInherit(connection.getIrodsAccount().getZone(), landingZoneInvestigationPath, true);
        // Enable inheritance for project data folder
        collectionAO.setAccessPermissionInherit(connection.getIrodsAccount().getZone(), landingZoneProjectPath + "/data", true);
        // Create data directory under investigation
        connection.getFileFactory().instanceIRODSFile(landingZoneInvestigationPath + "/data").mkdirs();
        // Set write permission for data folder
        // collectionAO.setAccessPermissionWrite(connection.getIrodsAccount().getZone(), landingZoneInvestigationPath + "/data", groupName, true);

        for (Person person : allResearcher) {
            String mbox = person.getEmail().toLowerCase().replaceAll("mailto:", "");
            try {
                userAO.findByName(mbox);
            } catch (DataNotFoundException e) {
                logger.info("Creating user: " + mbox);
                User user = new User();
                user.setName(mbox);
                user.setUserType(UserTypeEnum.RODS_USER);
                userAO.addUser(user);
            }

            logger.info("Adding user " + mbox + " to group " + ug.getUserGroupName());
            boolean inProjectGroup = false;
            for (User listUserGroupMember : userGroupAO.listUserGroupMembers(groupName)) {
                if (listUserGroupMember.getName().contains(mbox)) {
                    inProjectGroup = true;
                }
            }

            if (!inProjectGroup) {
                userGroupAO.addUserToGroup(groupName, mbox, null);
            }

            boolean inUsersGroup = false;
            for (User listUserGroupMember : userGroupAO.listUserGroupMembers("users")) {
                if (listUserGroupMember.getName().contains(mbox)) {
                    inUsersGroup = true;
                    break;
                }
            }

            if (!inUsersGroup) {
                userGroupAO.addUserToGroup("users", mbox, null);
            }
        }
    }

    public void logging(String message) {
        System.err.println(message);
        logger.info(message);
        this.message = this.message + "\n" + message;
    }

    private void setPrefixes() {
        Scanner scanner = new Scanner(WebGeneric.getFileContentFromResources("queries/prefixes.txt"));
        while (scanner.hasNextLine()) {
            String line = scanner.nextLine();
            domain.getRDFSimpleCon().setNsPrefix(line.split(" ")[1].replaceAll(":", ""), line.split(" ")[2].replaceAll("[<>]", ""));
        }
    }

    /**
     * Validation is an on going process but the Empusa API should take care of most things...
     */
    private void validate() throws Exception {

        // Check if duplicate identifiers have been used in the project
        Iterator<ResultLine> resultLineIterator = domain.getRDFSimpleCon().runQuery("getAllIdentifiers.txt", true).iterator();
        boolean failed = false;

        while (resultLineIterator.hasNext()) {
            ResultLine resultLine = resultLineIterator.next();
            String identifier = resultLine.getLitString("identifier");
            // Check for illegal characters for now it is only a "/"
            boolean allowed = identifier.matches("[A-Za-z0-9\\-_.]+");
            if (!allowed) {

                if (identifier.contains(" ")) {
                    logging("Identifier \"" + identifier + "\" contains a space, only alphanumeric and - _ : . are allowed");
                } else {
                    logging("Identifier \"" + identifier + "\" contains an illegal character, only alphanumeric and - _ : . are allowed");
                }

            }

            // Check for identifier counts
            int counts = resultLine.getLitInt("count");
            if (counts > 1) {
                failed = true;
                logging("Identifier " + identifier + " used in multiple places at least " + counts + " times");
            }
        }

        if (failed && ApplicationServiceInitListener.debug) {
            domain.save("multipleidentifiers.ttl");
            throw new Exception();
        }

        if (ApplicationServiceInitListener.debug) {
            domain.save("debug.ttl");
        }

        // Validates if all assays have the same name
        resultLineIterator = domain.getRDFSimpleCon().runQuery("getAssayTitles.txt", true).iterator();
        while (resultLineIterator.hasNext()) {
            ResultLine resultLine = resultLineIterator.next();
            if (resultLine.getLitInt("count") > 1) {
                resultLine.getLitString("desc");
                resultLine.getLitString("title");
                logging("It is advised that title and description is unique for each assay (e.g. Amplicon data from chicken sampled on day 30 from Farm 1)");
                // throw new Exception("Multiple usage of title and description detected in the assay descriptions");
            }
        }

        // Validate barcodes if they are used only once per library (prevents copy paste errors)
        resultLineIterator = domain.getRDFSimpleCon().runQuery("getBarcodeValidation.txt", true).iterator();
        while (resultLineIterator.hasNext()) {
            ResultLine resultLine = resultLineIterator.next();
            // Only enters this area when the same barcode is used multiple times for the same library
            logging("The same barcode " + resultLine.getLitString("fwd") + " " + resultLine.getLitString("rev") + " used multiple times in the library " + resultLine.getIRI("library"));
            throw new Exception("Multiple usage of the barcode detected");
        }

        // Validate amplicon assays if they have only 2 datasets when they are still multiplexed
        resultLineIterator = domain.getRDFSimpleCon().runQuery("getLibraryCount.txt", true).iterator();
        if (resultLineIterator.hasNext()) {
            ResultLine resultLine = resultLineIterator.next();
            logging("One of the assays in library name " + resultLine.getLitString("libraryName") + " is associated with the wrong library");
            logging("The following library files are involved");
            while (resultLineIterator.hasNext()) {
                resultLine = resultLineIterator.next();
                logging(resultLine.getLitString("filename"));
            }
            // Only enters this area when the same barcode is used multiple times for the same library
            throw new Exception("Wrong library association detected");
        }

        // Final validation

        // Obtain all Investigation, Study, Observation Units, Assays, Samples
        HashMap<String, ArrayList<String>> store = new HashMap<>();

        // Get all the projects
        ArrayList<String> projects = new ArrayList<>();
        domain.getRDFSimpleCon().runQuery("getProjects.txt", true).iterator().
                forEachRemaining(resultLine -> projects.add(resultLine.getIRI("iri")));
        logger.info("Detected " + projects.size() + " investigations");
        store.put("projects", projects);

        // Get all the investigations
        ArrayList<String> investigations = new ArrayList<>();
        domain.getRDFSimpleCon().runQuery("getInvestigations.txt", true).iterator().
                forEachRemaining(resultLine -> investigations.add(resultLine.getIRI("iri")));
        logger.info("Detected " + investigations.size() + " investigations");
        store.put("investigations", investigations);

        // Get all the studies
        ArrayList<String> studies = new ArrayList<>();
        domain.getRDFSimpleCon().runQuery("getStudies.txt", true).iterator().
                forEachRemaining(resultLine -> studies.add(resultLine.getIRI("iri")));
        logger.info("Detected " + studies.size() + " studies");
        store.put("studies", studies);

        // Get all the observation units
        ArrayList<String> observationUnits = new ArrayList<>();
        domain.getRDFSimpleCon().runQuery("getObservationUnits.txt", true).iterator().
                forEachRemaining(resultLine -> observationUnits.add(resultLine.getIRI("iri")));
        logger.info("Detected " + observationUnits.size() + " observationunits");
        store.put("observationunits", observationUnits);

        // Get all the samples
        ArrayList<String> samples = new ArrayList<>();
        domain.getRDFSimpleCon().runQuery("getSamples.txt", true).iterator().
                forEachRemaining(resultLine -> samples.add(resultLine.getIRI("sample")));
        logger.info("Detected " + samples.size() + " samples");
        store.put("samples", samples);

        Integer sum = store.values().stream().mapToInt(List::size).sum();
        logging("Total number of objects " + sum);
        // Perform query to get the full connection
        domain.getRDFSimpleCon().runQuery("getAllConnections.txt", true).iterator().
                forEachRemaining(resultLine -> {
                    store.get("projects").remove(resultLine.getIRI("project"));
                    store.get("investigations").remove(resultLine.getIRI("investigation"));
                    store.get("studies").remove(resultLine.getIRI("study"));
                    store.get("observationunits").remove(resultLine.getIRI("observationUnit"));
                    // store.get("assay").remove(resultLine.getIRI("assay"));
                    store.get("samples").remove(resultLine.getIRI("sample"));
                });

        ArrayList<String> results = new ArrayList<>();
        store.values().forEach(results::addAll);

        if (results.size() > 0) {
            ArrayList<String> sampleList = store.get("samples");
            if (sampleList.size() > 0) {
                if (sampleList.size() > 5) {
                    sampleList = new ArrayList<>(sampleList.subList(0, 4));
                    sampleList.add("... more samples have been found");
                }
                logging("Not all samples are coupled to assays which could be intentionally, please check.");
                for (String sample : sampleList) {
                    logging(sample);
                }
            }

            if (results.size() != store.get("samples").size()) {
                for (String result : results) {
                    logging("Mismatch with " + result + "... Indicating a typo in the identifiers between the different sheets");
                }

                throw new Exception("Mismatch with " + StringUtils.join(store.values(), " "));
            }
        } else {
            logging("All identifiers are accounted for saving the results");
        }

        // Double check with ontology for properties that are obligatory
        resultLineIterator = domain.getRDFSimpleCon().runQuery("getProjects.txt", true).iterator();
        while (resultLineIterator.hasNext()) {
            Project project = (Project) domain.getObject(resultLineIterator.next().getIRI("iri"), Project.class);
            project.getAllResearcher().forEach(person -> person.getGivenName());
            project.getTitle();
            project.getDescription();
            project.getIdentifier();
            for (Investigation investigation : project.getAllHasPart()) {
                for (Study study : investigation.getAllHasPart()) {
                    for (observation_unit observation_unit : study.getAllHasPart()) {
                        for (Sample sample : observation_unit.getAllHasPart()) {
                            for (Assay assay : sample.getAllHasPart()) {
//                                if (assay.getClassTypeIri().endsWith("AmpliconLibraryAssay")) {
//                                    AmpliconLibraryAssay assay1 = (AmpliconLibraryAssay) assay;
//                                    for (Data_sample data_sample : assay1.getAllHasPart()) {
//                                        data_sample.getName();
//                                    }
//                                } else if (assay.getClassTypeIri().endsWith("AmpliconAssay")) {
//                                    AmpliconAssay assay1 = (AmpliconAssay) assay;
//                                    for (Data_sample data_sample : assay1.getAllHasPart()) {
//                                        data_sample.getName();
//                                    }
//                                } else if (assay.getClassTypeIri().endsWith("RNASeqAssay")) {
//                                    RNASeqAssay assay1 = (RNASeqAssay) assay;
//                                    for (Data_sample data_sample : assay1.getAllHasPart()) {
//                                        data_sample.getName();
//                                    }
//                                } else if (assay.getClassTypeIri().endsWith("DNASeqAssay")) {
//                                    DNASeqAssay assay1 = (DNASeqAssay) assay;
//                                    for (Data_sample data_sample : assay1.getAllHasPart()) {
//                                        data_sample.getName();
//                                    }
//                                } else
                                    if (assay.getClassTypeIri().endsWith("Assay")) {
                                    for (Data_sample data_sample : assay.getAllHasPart()) {
                                        data_sample.getName();
                                    }
                                } else {
                                    throw new Exception("Unknown type " + assay.getClassTypeIri());
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Sample creation module from the sample sheet
     *
     * @param sheet
     * @param metadataArrayList
     * @param samplePackage
     */
    private void sampleCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList, String samplePackage) throws Exception {
        logger.info("Parsing sample sheet");
        boolean headerRow = true;

        Set<String> parsedHeaders = new HashSet<>();

        // To store the header of the excel sheet in
        ArrayList<String> header = new ArrayList<>();

        for (Row row : sheet) {
            if (row.getRowNum() % 100 == 0)
                logger.debug("Processing row " + row.getRowNum());
            // Default row parser...
            // First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);
                    if (status) {
                        logging("Please complete the missing cell values");
                        throw new Exception("");
                    }
                    Sample sample = domain.make(Sample.class, PREFIX + "sample/" + row.getCell(header.indexOf(SAMPLE_IDENTIFIER)).getStringCellValue().trim());
                    // sample.setIdentifier(iriCorrector(getStringValueCell(row.getCell(header.indexOf(SAMPLE_IDENTIFIER)))));
                    // TODO temp fix as schema.org name is preferred but metadata sheet also expects sample name from mixs
                    // sample.setName(iriCorrector(getStringValueCell(row.getCell(header.indexOf("sample name")))));
                    samplePackage = StringUtils.capitalize(samplePackage.replaceAll("-", "_").toLowerCase());
                    logger.debug("Sample package: " + samplePackage);

                    try {
                        sample.setSamplePackage(Mixs.valueOf(samplePackage));
                    } catch (Exception e) {
                        sample.setSamplePackage(Mixs.Unknown);
                    }

                    // parsedHeaders.add("Sample identifier");

                    // Creating the observation unit...
                    String observationUnitIdentifier = getStringValueCell(row.getCell(header.indexOf(OBSERVATION_UNIT_IDENTIFIER)));
                    observation_unit observation_unit = domain.make(org.purl.ppeo.PPEO.owl.domain.observation_unit.class, PREFIX + "ObservationUnit/OU_" + observationUnitIdentifier);
                    // observation_unit.setIdentifier(iriCorrector(getStringValueCell(row.getCell(header.indexOf(OBSERVATION_UNIT_IDENTIFIER)))));// .getStringCellValue().trim()));

                    // Parse all other headers to sample
                    // parsedHeaders.addAll(Arrays.asList(OBSERVATION_UNIT_IDENTIFIER));
                    otherColumns(header, parsedHeaders, row, sample.getResource().getURI(), metadataArrayList);
                    // Link sample to OU and validate
                    observation_unit.addHasPart(sample);
                    // Identifier required for other queries
                    observation_unit.setIdentifier(observationUnitIdentifier);
                }
            }
        }
    }

    /**
     * Checks if the obligatory fields are filled in for each row
     *
     * @param header            the specific column / header being requested
     * @param row               the row being parsed
     * @param metadataArrayList is a list of metadata objects
     * @return
     */
    private boolean checkObligatoryFields(ArrayList<String> header, Row row, ArrayList<Metadata> metadataArrayList) {
        boolean status = false;

        String identifier = row.getCell(0).getStringCellValue().trim();
        for (Metadata metadata : metadataArrayList) {
            // Non obligatory elements are skipped
            if (!metadata.getRequirement().toLowerCase().contains("m")) continue;
            // Obtain name of metadata element
            if (header.indexOf(metadata.getLabel()) < 0) {
                logging("in sheet " + row.getSheet().getSheetName() + " row " + row.getRowNum() + " " + identifier + " does not contain column " + metadata.getLabel() + " which is obligatory");
                status = true;
            } else {
                CellType cellType = row.getCell(header.indexOf(metadata.getLabel()), Row.MissingCellPolicy.CREATE_NULL_AS_BLANK).getCellType();
                if (cellType.equals(CellType.BLANK)) {
                    logging("in sheet " + row.getSheet().getSheetName() + " row " + row.getRowNum() + " " + identifier + " contains an empty cell for column " + metadata.getLabel() + " which is obligatory");
                    status = true;
                }
            }
        }
        return status;
    }

    /**
     * Corrects URLS which could contain a space due to naming conventions, all should eventually be replaced by PID's
     *
     * @param iriPart part of the IRI that needs to be corrected
     * @return corrected IRI
     */
    private String iriCorrector(String iriPart) {
        return iriPart.trim().replaceAll("[ \\(\\)]", "_").replaceAll(" +", "_").replaceAll("_+", "_");
    }

    /**
     * @param sheet
     * @param metadataArrayList
     * @param packageName
     * @throws Exception
     */
    private void assayCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList, String packageName) throws Exception {
        logger.info("Parsing " + sheet.getSheetName());
        ArrayList<String> header = new ArrayList<>();

        // Assay type after " - " split, strip, camel case it and then remove spaces and turn into a URI
        String assayTypeIRI;
        if (sheet.getSheetName().contains(" - ")) {
            String[] name = sheet.getSheetName().split(" - ");
            String assayName = name[name.length-1];
            boolean validAssayName = false;
            // Check if it is part of the excel package
            for (Metadata metadata : metadataArrayList) {
                if (metadata.getPackage().toLowerCase(Locale.ROOT).contains(assayName.toLowerCase(Locale.ROOT))) {
                    validAssayName = true;
                    break;
                }
            }
            if (!validAssayName) {
                throw new Exception("This assay name is not supported: " + assayName);
            }

            // Create IRI
            assayName = StringUtils.capitalize(assayName).replaceAll(" +", "");
            assayTypeIRI = "http://m-unlock.nl/ontology/" + assayName;

        } else {
            // some kind of exception
            throw new Exception("Assay sheet formatting is not correct should be according to: Assay - type");
        }

        // Header capture
        boolean headerRow = true;
        // Iterating over each row
        for (Row row : sheet) {
            if (row.getRowNum() % 100 == 0) {
                logger.info("Processing row " + row.getRowNum());
            }
            // First cell has to be a string... and due to numeric identifiers sometimes it will be parsed incorrectly
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            // Standard row/cell check
            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);

                    if (status) {
                        // Logging from validate obligatory contains information that have been passed to the interface
                        throw new Exception("");
                    }

                    // If (forward) filename not given, assume this is ok as the files might not be known yet...
                    Cell fileCell = null;
                    if (header.indexOf(FORWARD_FILENAME) != -1) {
                        fileCell = row.getCell(header.indexOf(FORWARD_FILENAME), Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);
                    } else if (header.indexOf(FILE_NAME) != -1) {
                        fileCell = row.getCell(header.indexOf(FILE_NAME), Row.MissingCellPolicy.CREATE_NULL_AS_BLANK);
                    }

                    // TODO Reduce list to the core elements
                    Set<String> parsedHeaders = new HashSet<>();

                    // Obtain identifier
                    int index = header.indexOf(ASSAY_IDENTIFIER);
                    String assayIdentifier = row.getCell(index).getStringCellValue().trim();

                    /**
                     * Specific sequence assay information
                     */

                    // Check if assay identifier exists?
                    if (domain.getRDFSimpleCon().runQuery("getAssay.txt", true, assayIdentifier).iterator().hasNext()) {
                        // Duplication detected, check if SameAs is present
                        if (header.indexOf(SAME_AS) == -1) {
                            throw new DuplicateDataException("Assay with identifier " + assayIdentifier + " already exists!\n" +
                                    "Check for duplicate row entry in your input file");
                        } else {
                            // Further, testing if it is of the same type
                            // assayType
                            Iterator<ResultLine> resultLines = domain.getRDFSimpleCon().runQuery("getAssay.txt", true, assayIdentifier).iterator();
                            while (resultLines.hasNext()) {
                                ResultLine resultLine = resultLines.next();
                                String type = resultLine.getIRI("type");

                                if (!type.toLowerCase().endsWith(sheet.getSheetName().toLowerCase())) {
                                    throw new DuplicateDataException("Assay with identifier " + assayIdentifier + " already exists as a different assay type!\n" +
                                            "Check for duplicate identifiers in your assay sheets");
                                }
                            }
                            if (row.getCell(header.indexOf(SAME_AS)) == null) {
                                throw new Exception("Identifier " + assayIdentifier + " has a duplicate entry but no value for the SameAs column.");
                            }
                            String sameAsValue = getStringValueCell(row.getCell(header.indexOf(SAME_AS))).trim();
                            // If identifier is not the same throw an error
                            if (!assayIdentifier.equals(sameAsValue)) {
                                logging("SameAs detected but assay identifier not the same as sameAs identifier " + assayIdentifier + " and " + sameAsValue);
                                throw new DuplicateDataException("SameAs detected but assay identifier not the same as sameAs identifier " + assayIdentifier + " and " + sameAsValue);
                            } else {
                                // Duplicate assay with same identifier detected
                                parsedHeaders.add("Same as");
                            }
                        }
                    }

                    // Setting the assay URL based on the type and identifier
                    String ASSAYIRI = PREFIX + assayTypeIRI.replaceAll(".*/", "") + "/ASY_" + row.getCell(index).getStringCellValue().trim() + "";

                    // Generic place holders
                    Assay assay = domain.make(Assay.class, ASSAYIRI);
                    assay.addAdditionalType(assayTypeIRI);

//                    // Section for amplicon / library data
//                    if (assayType.equals(AssayType.Amplicon)) {
//                        AmpliconAssay ampliconAssay = domain.make(AmpliconAssay.class, ASSAYIRI);
//                        assay = ampliconAssay;
//                    } else if (assayType.equals(AssayType.AmpliconLibrary)) {
//                        // Create amplicon library object
//                        if (header.indexOf(LIBRARY_NAME) == -1) {
//                            logging("Missing LibraryName for AmpliconLibraryAssay");
//                        }
//                        String assayLibraryIRI = PREFIX + AssayType.AmpliconLibrary.getIRI().replaceAll(".*/", "") + "/" + getStringValueCell(row.getCell(header.indexOf(LIBRARY_NAME))).trim();
//                        String assayIRI = PREFIX + AssayType.Amplicon.getIRI().replaceAll(".*/", "") + "/" + getStringValueCell(row.getCell(header.indexOf(ASSAY_IDENTIFIER))).trim();
//
//                        // logger.debug("Library " + assayLibraryIRI);
//
//                        AmpliconLibraryAssay ampliconLibraryAssay = domain.make(AmpliconLibraryAssay.class, assayLibraryIRI);
//                        String ampliconLibraryAssayIdentifier = getStringValueCell(row.getCell(header.indexOf(LIBRARY_NAME))).trim();
//                        ampliconLibraryAssay.setIdentifier(ampliconLibraryAssayIdentifier);
//                        ampliconLibraryAssay.setDescription("Amplicon library " + ampliconLibraryAssayIdentifier);
//                        ampliconLibraryAssay.setLogicalPath("null");
//
//                        // Cell can be numeric
////                         Sample sample = getSample(iriCorrector(getStringValueCell(row.getCell(header.indexOf("SampleIdentifier")))));
//                        // TODO check if we should do this? Amplicon assay should be a library and attached to assay of type amplicon
//                        // sample.addAssay(ampliconLibraryAssay);
//                        // ampliconLibraryAssay.setSample(sample);
//                        // parsedHeaders.add(LIBRARY_NAME);
//                        // Skip identifier addition due to amplicon assays having many different identifiers for the same library
//                        parsedHeaders.add(ASSAY_IDENTIFIER);
//                        parsedHeaders.add(ASSAY_DESCRIPTION);
//                        // Prevent adding file objects as strings
//                        parsedHeaders.add(FORWARD_FILENAME);
//                        parsedHeaders.add(REVERSE_FILENAME);
//
//                        otherColumns(header, parsedHeaders, row, ampliconLibraryAssay.getResource().getURI(), metadataArrayList);
//                        // Remove this for further processing of the assay
//                        parsedHeaders.removeAll(List.of(new String[]{ASSAY_IDENTIFIER, ASSAY_DESCRIPTION}));
//
//                        // Create amplicon object even though its not demultiplexed yet?...
//                        AmpliconAssay ampliconAssay = domain.make(AmpliconAssay.class, assayIRI);
//                        // Make link to the amplicon library
//                        ampliconAssay.setLibrary(ampliconLibraryAssay);
//                        assay = ampliconAssay;
//
//                        // Section for RNA sequencing data
//                    } else if (assayType.equals(AssayType.Transcriptomics)) {
//                        RNASeqAssay rnaSeqAssay = domain.make(RNASeqAssay.class, ASSAYIRI);
//                        assay = rnaSeqAssay;
//                        // Section for DNA sequencing data
//                    } else if (assayType.equals(AssayType.Genomics)) {
//                        DNASeqAssay dnaSeqAssay = domain.make(DNASeqAssay.class, ASSAYIRI);
//                        assay = dnaSeqAssay;
//                    } else {
//                        logger.info("Created default Assay");
//                        assay = domain.make(Assay.class, ASSAYIRI);
//                        // TODO add additional type based on sheet name and default IRI
//                    }

                    /**
                     * Generic Assay information
                     */

                    // assay.setDescription(getStringValueCell(row.getCell(header.indexOf(ASSAY_DESCRIPTION))));
                    // assay.setIdentifier(iriCorrector(row.getCell(header.indexOf(ASSAY_IDENTIFIER)).getStringCellValue().trim()));

                    // parsedHeaders.addAll(Arrays.asList(ASSAY_DESCRIPTION, ASSAY_TITLE, ASSAY_IDENTIFIER, PLATFORM));

                    if (fileCell.getCellType().equals(CellType.BLANK)) {
                        logger.debug("No file found for " + assay.getResource().getURI());
                    } else {
                        String fileName = getStringValueCell(fileCell);
                        // Check the FileNameForwardPath and FileNameReversePath first
                        boolean paired = header.contains(REVERSE_FILENAME) && row.getCell(header.indexOf(REVERSE_FILENAME)) != null && row.getCell(header.indexOf(REVERSE_FILENAME)).getStringCellValue().trim().length() > 0;

                        if (paired) {
                            PairedSequenceDataSet pairedSequenceDataSetForward = domain.make(PairedSequenceDataSet.class, PREFIX + fileName);
                            pairedSequenceDataSetForward.setName(new File(fileName).getName().trim());
                            pairedSequenceDataSetForward.setContentUrl("null");
                            pairedSequenceDataSetForward.setBase64("null");
                            pairedSequenceDataSetForward.setSha256("null");
                            pairedSequenceDataSetForward.setReadLength(0L);
                            pairedSequenceDataSetForward.setReads(0L);
                            pairedSequenceDataSetForward.setBases(0L);
                            pairedSequenceDataSetForward.setContentSize(0L);
                            pairedSequenceDataSetForward.setFileFormat(setFileFormat(pairedSequenceDataSetForward.getName()));
                            pairedSequenceDataSetForward = (PairedSequenceDataSet) setSequencePlatform(pairedSequenceDataSetForward, header, row);

                            String fileNameReverse = row.getCell(header.indexOf(REVERSE_FILENAME)).getStringCellValue().trim();
                            if (fileNameReverse.contains("Enter only when data is demultiplexed"))
                                fileNameReverse = "";

                            if (fileNameReverse.length() > 0) {
                                PairedSequenceDataSet pairedSequenceDataSetReverse = domain.make(PairedSequenceDataSet.class, PREFIX + fileNameReverse);
                                pairedSequenceDataSetReverse.setName(new File(fileNameReverse).getName().trim());
                                pairedSequenceDataSetReverse.setContentUrl("null");
                                pairedSequenceDataSetReverse.setBase64("null");
                                pairedSequenceDataSetReverse.setSha256("null");
                                pairedSequenceDataSetReverse.setReadLength((long) 0);
                                pairedSequenceDataSetReverse.setContentSize(0L);
                                pairedSequenceDataSetReverse.setReads((long) 0);
                                pairedSequenceDataSetReverse.setBases((long) 0);

                                pairedSequenceDataSetReverse.setFileFormat(setFileFormat(pairedSequenceDataSetReverse.getName()));
                                pairedSequenceDataSetReverse = (PairedSequenceDataSet) setSequencePlatform(pairedSequenceDataSetReverse, header, row);

                                // Cross linking requires validator to be temporarily disabled
                                domain.disableCheck();
                                pairedSequenceDataSetReverse.setPaired(pairedSequenceDataSetForward);
                                domain.enableCheck();
                                pairedSequenceDataSetForward.setPaired(pairedSequenceDataSetReverse);
                                assay.addHasPart(pairedSequenceDataSetForward);
                                assay.addHasPart(pairedSequenceDataSetReverse);

                                parsedHeaders.addAll(Arrays.asList(FORWARD_FILENAME, REVERSE_FILENAME));
                            }
                            // If it is one of the sequence assay types
                        } else if (!assay.getClassTypeIri().endsWith("/Assay")) {
                            SingleSequenceDataSet singleSequenceDataSet = domain.make(SingleSequenceDataSet.class, PREFIX + fileName);
                            singleSequenceDataSet.setName(new File(fileName).getName().trim());
                            singleSequenceDataSet.setContentUrl("null");
                            singleSequenceDataSet.setBase64("null");
                            singleSequenceDataSet.setSha256("null");
                            singleSequenceDataSet.setReadLength(0L);
                            singleSequenceDataSet.setContentSize(0L);
                            singleSequenceDataSet.setReads(0L);
                            singleSequenceDataSet.setBases(0L);

                            singleSequenceDataSet.setFileFormat(setFileFormat(singleSequenceDataSet.getName()));
                            singleSequenceDataSet = (SingleSequenceDataSet) setSequencePlatform(singleSequenceDataSet, header, row);

                            assay.addHasPart(singleSequenceDataSet);

                            parsedHeaders.addAll(Arrays.asList(FORWARD_FILENAME));
                        } else {
                            Data_sample data_sample = domain.make(Data_sample.class, PREFIX + fileName);
                            data_sample.setName(new File(fileName).getName().trim());
                            data_sample.setContentUrl("null");
                            data_sample.setBase64("null");
                            data_sample.setSha256("null");
                            data_sample.setContentSize(0L);
                            data_sample.setFileFormat(setFileFormat(data_sample.getName()));
                            assay.addHasPart(data_sample);
                        }
                    }

                    assay.setLogicalPath("null");
//                    assay.setPhysicalPath("null");

                    // Creating the observation unit which was previously a study...
//                        org.purl.ppeo.PPEO.owl.domain.observation_unit observation_unit = domain.make(org.purl.ppeo.PPEO.owl.domain.observation_unit.class, PREFIX + "ObservationUnit/OU_" + row.getCell(header.indexOf("ObservationUnitIdentifier")).getStringCellValue().trim());
//
//                        observation_unit.setIdentifier(iriCorrector(row.getCell(header.indexOf("ObservationUnitIdentifier")).getStringCellValue().trim()));
//                        parsedHeaders.addAll(Arrays.asList("ObservationUnitIdentifier"));

                    // Get sample
                    Sample sample = getSample(getStringValueCell(row.getCell(header.indexOf(SAMPLE_IDENTIFIER))));
                    // parsedHeaders.addAll(Arrays.asList(SAMPLE_IDENTIFIER));

                    // assay is still an amplicon assay but if it has an amplicon library we should move the files from amplicon assay to amplicon library assay
//                    if (assayType.equals(AssayType.AmpliconLibrary)) {
//                        AmpliconAssay ampliconAssay = (AmpliconAssay) assay;
//                        AmpliconLibraryAssay library = ampliconAssay.getLibrary();
//                        for (Data_sample data_sample : ampliconAssay.getAllHasPart()) {
//                            library.addHasPart(data_sample);
//                            ampliconAssay.remHasPart(data_sample);
//                        }
//                    }

                    // Perform reference incorporation
                    if (header.indexOf(REFERENCE) != -1) {
                        parsedHeaders.addAll(Arrays.asList(REFERENCE));
                        // Check if reference has value
                        if (row.getCell(header.indexOf(REFERENCE)) != null) {
                            String referenceName = row.getCell(header.indexOf(REFERENCE)).getStringCellValue().trim();
                            Data_sample dataSample = domain.make(Data_sample.class, PREFIX + referenceName);
                            dataSample.setName(referenceName.trim());
                            dataSample.setSha256("null");
                            dataSample.setBase64("null");
                            dataSample.setContentUrl("null");
                            dataSample.setContentSize(0L);
                            dataSample.setFileFormat(setFileFormat(dataSample.getName()));
                            assay.setReference(dataSample);
                        }
                    }
                    // Parse all other columns
                    otherColumns(header, parsedHeaders, row, assay.getResource().getURI(), metadataArrayList);
                    // Coupling the assay to the sample
                    sample.addHasPart(assay);
                }
            }
        }
    }

    private boolean isBlank(Row row, int columnIndex) {
        return row.getCell(columnIndex, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK).getCellType().equals(CellType.BLANK);
    }

    private SequenceDataSet setSequencePlatform(SequenceDataSet sequenceDataSet, ArrayList<String> header, Row row) throws Exception {
        if (header.indexOf(PLATFORM) == -1) {
            logging("SequencingPlatform column is missing in the excel sheet");
            throw new Exception("");
        }

        String seqPlatform = row.getCell(header.indexOf(PLATFORM)).getStringCellValue().trim().toLowerCase();

        if (seqPlatform.toLowerCase().contains("illumina")) {
            sequenceDataSet.setSeqPlatform(Illumina);
            return sequenceDataSet;
        }

        if (seqPlatform.toLowerCase().contains("pacbio")) {
            sequenceDataSet.setSeqPlatform(PacBio);
            return sequenceDataSet;
        }

        if (seqPlatform.toLowerCase().contains("nanopore")) {
            sequenceDataSet.setSeqPlatform(Nanopore);
            return sequenceDataSet;
        }

        logging("Failed to identify platform (illumina, pacbio, nanopore)");
        throw new Exception("");
    }


    /**
     * @param sampleIdentifier
     * @return
     * @throws Exception
     */
    private Sample getSample(String sampleIdentifier) throws Exception {
        try {
            return domain.make(Sample.class, domain.getRDFSimpleCon().runQuerySingleRes("getSample.txt", false, sampleIdentifier).getIRI("sample"));
        } catch (Exception e) {
            // Get all observed units
            domain.getRDFSimpleCon().runQuery("getSamples.txt", true).iterator().forEachRemaining(resultLine -> logger.debug("Samples available: " + resultLine.getIRI("sample")));
            domain.save("getSamples.ttl");
            throw new Exception("Cannot find sample with identifier: >" + sampleIdentifier + "<");
        }
    }

    private observation_unit getObservationUnit(String observationUnitIdentifier) throws Exception {
        try {
            return domain.make(observation_unit.class, domain.getRDFSimpleCon().runQuerySingleRes("getObservationUnit.txt", false, observationUnitIdentifier).getIRI("observationUnit"));
        } catch (Exception e) {
            // Get all observed units
            domain.getRDFSimpleCon().runQuery("getObservationUnits.txt", true).iterator().forEachRemaining(resultLine -> logging("Observation units available: " + resultLine.getLitString("observationUnit")));
            domain.save("issue.ttl");
            exception("Cannot find observation unit with identifier: >" + observationUnitIdentifier + "< perhaps no sample has been registered for this observation unit?");
            return null;
        }
    }

    private void exception(String exception) throws Exception {
        message = message + "\n" + exception;
        throw new Exception(exception);
    }

    private Study getStudy(String studyIdentifier) throws Exception {
        try {
            return domain.make(Study.class, domain.getRDFSimpleCon().runQuerySingleRes("getStudy.txt", false, studyIdentifier).getIRI("study"));
        } catch (Exception e) {
            // Get all studies
            domain.getRDFSimpleCon().runQuery("getStudies.txt", true).iterator().forEachRemaining(resultLine -> logging("Studies available: " + resultLine.getLitString("observationUnit")));
            exception("Cannot find study with identifier: >" + studyIdentifier + "< assuming a mismatch with the observation unit");
            return null;
        }
    }

    private Investigation getInvestigation(String investigationIdentifier) throws Exception {
        try {
            return domain.make(Investigation.class, domain.getRDFSimpleCon().runQuerySingleRes("getInvestigation.txt", false, investigationIdentifier).getIRI("investigation"));
        } catch (Exception e) {
            // Get all studies
            domain.getRDFSimpleCon().runQuery("getInvestigations.txt", true).iterator().forEachRemaining(resultLine -> logging("Investigations available: " + resultLine.getLitString("investigation")));
            exception("Cannot find investigation with identifier: >" + investigationIdentifier + "< assuming a mismatch with the study");
            return null;
        }
    }

    /**
     * Acquire project identifier which is created at the investigation sheet
     *
     * @param projectIdentifier the identifier for the project
     * @return the project identifier
     * @throws Exception
     */
    private Project getProject(String projectIdentifier) throws Exception {
        try {
            return domain.make(Project.class, domain.getRDFSimpleCon().runQuerySingleRes("getProject.txt", false, projectIdentifier).getIRI("project"));
        } catch (Exception e) {
            // Get all studies
            domain.getRDFSimpleCon().runQuery("getProjects.txt", true).iterator().forEachRemaining(resultLine -> logging("Projects available: " + resultLine.getLitString("project")));
            exception("Cannot find project with identifier: >" + projectIdentifier + "< assuming a mismatch with the investigation");
            return null;
        }
    }

    /**
     * @param sheet The observation unit sheet
     * @throws Exception
     */
    private void observationUnitCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList) throws Exception {
        logging("Observation unit creation");

        boolean headerRow = true;
        ArrayList<String> header = new ArrayList<>();
        Set<String> parsedHeaders = new HashSet<>();
        parsedHeaders.addAll(Arrays.asList(OBSERVATION_UNIT_IDENTIFIER, STUDY_IDENTIFIER));

        for (Row row : sheet) {
            // Default row parser... First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);
                    if (status) {
                        logging("Please complete the missing cell values");
                        throw new Exception("");
                    }

                    observation_unit observationUnit = getObservationUnit(iriCorrector(row.getCell(header.indexOf(OBSERVATION_UNIT_IDENTIFIER)).getStringCellValue().trim()));

                    // observationUnit.setDescription(getStringValueCell(row.getCell(header.indexOf("ObservationUnitDescription"))));//.getStringCellValue().trim());
                    // observationUnit.setIdentifier(iriCorrector(getStringValueCell(row.getCell(header.indexOf(OBSERVATION_UNIT_IDENTIFIER)))));//.getStringCellValue().trim()));

                    otherColumns(header, parsedHeaders, row, observationUnit.getResource().getURI(), metadataArrayList);

                    // Creating study section to map to observation unit
                    try {
                        Study study = domain.make(Study.class, PREFIX + "Study/STU_" + row.getCell(header.indexOf(STUDY_IDENTIFIER)).getStringCellValue().trim());
                        study.setIdentifier(iriCorrector(row.getCell(header.indexOf(STUDY_IDENTIFIER)).getStringCellValue().trim()));
                        observationUnit.setLogicalPath("null");
                        study.addHasPart(observationUnit);
                    } catch (Exception e) {
                        throw new Exception("Missing StudyIdentifier in row " + row.getRowNum() + " in excel sheet " + sheet.getSheetName());
                    }
                }
            }
        }
    }

    private void projectCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList) throws Exception {
        logging("Project creation");
//        RegexValidator regexValidator = new RegexValidator(metadataArrayList);

        boolean headerRow = true;
        ArrayList<String> header = new ArrayList<>();

        Set<String> parsedHeaders = new HashSet<>();
        // parsedHeaders.addAll(Arrays.asList(PROJECT_IDENTIFIER, PROJECT_DESCRIPTION, PROJECT_TITLE));

        for (Row row : sheet) {
            // Default row parser...
            // First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);
                    if (status) {
                        logging("Please complete the missing cell values");
                        throw new Exception("");
                    }

                    // Obtaining project information from investigation
                    Project project = getProject(iriCorrector(row.getCell(header.indexOf(PROJECT_IDENTIFIER)).getStringCellValue().trim()));

                    // project.setDescription(row.getCell(header.indexOf(PROJECT_DESCRIPTION)).getStringCellValue().trim());
                    // project.setTitle(row.getCell(header.indexOf(PROJECT_TITLE)).getStringCellValue().trim());
                    // project.setIdentifier(iriCorrector(row.getCell(header.indexOf(PROJECT_IDENTIFIER)).getStringCellValue().trim()));
                    // project.setLogicalPath("null");

                    otherColumns(header, parsedHeaders, row, project.getResource().getURI(), metadataArrayList);

                    // project.setIdentifier(iriCorrector(row.getCell(header.indexOf(PROJECT_IDENTIFIER)).getStringCellValue().trim()));

                    // otherColumns(header, parsedHeaders, row, project.getResource().getURI(), metadataArrayList);
                }
            }
        }
    }

    private void studyCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList) throws Exception {
        logging("Study creation");
//        RegexValidator regexValidator = new RegexValidator(metadataArrayList);

        boolean headerRow = true;
        ArrayList<String> header = new ArrayList<>();

        Set<String> parsedHeaders = new HashSet<>();
        parsedHeaders.addAll(Arrays.asList(STUDY_IDENTIFIER, STUDY_DESCRIPTION, STUDY_TITLE, INVESTIGATION_IDENTIFIER));

        for (Row row : sheet) {
            // Default row parser...
            // First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);
                    if (status) {
                        logging("Please complete the missing cell values");
                        throw new Exception("");
                    }

                    // Obtaining study from Observation Unit
                    Study study = getStudy(iriCorrector(row.getCell(header.indexOf(STUDY_IDENTIFIER)).getStringCellValue().trim()));

                    study.setDescription(row.getCell(header.indexOf(STUDY_DESCRIPTION)).getStringCellValue().trim());
                    study.setTitle(row.getCell(header.indexOf(STUDY_TITLE)).getStringCellValue().trim());
                    study.setIdentifier(iriCorrector(row.getCell(header.indexOf(STUDY_IDENTIFIER)).getStringCellValue().trim()));
                    study.setLogicalPath("null");
//                    study.setPhysicalPath("null");

                    Investigation investigation = domain.make(Investigation.class, PREFIX + "Investigation/INV_" + row.getCell(header.indexOf(INVESTIGATION_IDENTIFIER)).getStringCellValue().trim());
                    // investigation.getAllOwnedBy().forEach(element -> investigation.addOwnedBy(element));

                    otherColumns(header, parsedHeaders, row, study.getResource().getURI(), metadataArrayList);

                    investigation.setIdentifier(iriCorrector(row.getCell(header.indexOf(INVESTIGATION_IDENTIFIER)).getStringCellValue().trim()));
                    try {
                        investigation.addHasPart(study);
                    } catch (RuntimeException e) {
                        logging("Study " + study.getIdentifier() + " missing observationUnit perhaps the names do not match in the sheet?");
                        throw new RuntimeException(e);
                    }
                }
            }
        }
    }

    private boolean checkNotEmpty(Row row) {
        Iterator<Cell> rowIter = row.iterator();
        while (rowIter.hasNext()) {
            Cell cell = rowIter.next();
            if (!cell.getCellType().equals(CellType.BLANK))
                return true;
        }
        return false;
    }

    private String getProjectIdentifier(Sheet sheet) {
        logging("Get project identifier");
        boolean headerRow = true;
        ArrayList<String> header = new ArrayList<>();

        for (Row row : sheet) {
            // Default row parser...
            // First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    // Project identifier
                    int index = header.indexOf(PROJECT_IDENTIFIER);
                    return iriCorrector(row.getCell(index).getStringCellValue().trim().trim());
                }
            }
        }
        return null;
    }

    /**
     * @param sheet
     * @throws Exception
     */
    private void investigationCreation(Sheet sheet, ArrayList<Metadata> metadataArrayList) throws Exception {
        logging("Investigation creation");
//        RegexValidator regexValidator = new RegexValidator(metadataArrayList);

        boolean headerRow = true;
        ArrayList<String> header = new ArrayList<>();

        Set<String> parsedHeaders = new HashSet<>();
        parsedHeaders.addAll(Arrays.asList(INVESTIGATION_IDENTIFIER, INVESTIGATION_TITLE, INVESTIGATION_DESCRIPTION, "First name", "Last name", EMAIL, ORCID, ORGANIZATION, DEPARTMENT));

        for (Row row : sheet) {
            // Default row parser...
            // First cell has to be a string...
            if (row.getCell(0) != null && row.getCell(0).getCellType().equals(CellType.NUMERIC)) {
                row.getCell(0).setCellValue(String.valueOf(row.getCell(0).getNumericCellValue()));
            }

            if (row.getCell(0) != null && !row.getCell(0).getStringCellValue().trim().trim().startsWith("#") && row.getLastCellNum() != -1 && checkNotEmpty(row)) {
                if (headerRow) {
                    header = getHeader(row);
                    headerRow = false;
                } else {
                    boolean status = checkObligatoryFields(header, row, metadataArrayList);
                    if (status) {
                        logging("Please complete the missing cell values");
                        throw new Exception("");
                    }

                    // InvestigationIdentifier
                    Investigation investigation = getInvestigation(iriCorrector(row.getCell(header.indexOf(INVESTIGATION_IDENTIFIER)).getStringCellValue().trim().trim()));

                    if (investigation == null) {
                        throw new Exception("Investigation not matching to study, reason could be that there is no study with this investigation: " + row.getCell(header.indexOf(INVESTIGATION_IDENTIFIER)).getStringCellValue().trim().trim());
                    }

                    int index = header.indexOf(INVESTIGATION_IDENTIFIER);
                    String investigationIdentifier = iriCorrector(row.getCell(index).getStringCellValue().trim().trim());

                    investigation.setIdentifier(investigationIdentifier);
                    // InvestigationTitle	InvestigationDescription
                    investigation.setTitle(row.getCell(header.indexOf(INVESTIGATION_TITLE)).getStringCellValue().trim().trim());
                    try {
                        investigation.setDescription(row.getCell(header.indexOf(INVESTIGATION_DESCRIPTION)).getStringCellValue().trim().trim());
                    } catch (NullPointerException e) {
                        logging("Investigation is missing a description");
                    }

                    // FirstName	LastName	Email
                    // ToDO remove this from the final file due to person information?
                    Person person = domain.make(Person.class, PREFIX + "/person/" + DigestUtils.md5Hex(row.getCell(header.indexOf(EMAIL)).getStringCellValue().trim().trim()));
                    person.setGivenName(row.getCell(header.indexOf(FIRSTNAME)).getStringCellValue().trim().trim());
                    person.setFamilyName(row.getCell(header.indexOf(LASTNAME)).getStringCellValue().trim().trim());
                    person.setEmail("mailto:" + row.getCell(header.indexOf(EMAIL)).getStringCellValue().trim().trim());
                    person.setGivenName(person.getGivenName() + " " + person.getFamilyName());

                    // If orcid is available
                    if (header.indexOf(ORCID) >= 0 && row.getCell(header.indexOf(ORCID)) != null) {
                        String value = getStringValueCell(row.getCell(header.indexOf(ORCID))).trim().trim();
                        if (value.length() > 1)
                            person.setOrcid(value);
                    }

                    // TODO add property for depiction
                    person.setDepartment(row.getCell(header.indexOf(DEPARTMENT)).getStringCellValue().trim().trim());

                    Organization organization = domain.make(Organization.class, "http://m-unlock.nl/ontology/organization/" + row.getCell(header.indexOf(ORGANIZATION)).getStringCellValue().trim().trim().replaceAll(" ", "_"));
                    organization.setLegalName(row.getCell(header.indexOf(ORGANIZATION)).getStringCellValue().trim().trim());

                    person.addMemberOf(organization);

                    investigation.setLogicalPath("null");
                    investigation.addHasContributor(person);

                    // Check for other arguments
                    otherColumns(header, parsedHeaders, row, investigation.getResource().getURI(), metadataArrayList);

                    // Investigation attachment to project
                    Project project = domain.make(Project.class, "http://m-unlock.nl/ontology/project/" + row.getCell(header.indexOf(PROJECT_IDENTIFIER)).getStringCellValue().trim().trim());
                    project.setIdentifier(iriCorrector(row.getCell(header.indexOf(PROJECT_IDENTIFIER)).getStringCellValue().trim().trim()));
                    project.addHasPart(investigation);
                    project.addResearcher(person);
                }
            }
        }
    }

    /**
     * @param header            the header of the sheet
     * @param ignoredHeaders    headers that have already been parsed and can be ignored
     * @param row               the row that is being processed
     * @param iri               the base url used?
     * @param metadataArrayList the metadata objects for this particular class type e.g. assay, sample, investigation
     */
    private void otherColumns(ArrayList<String> header, Set<String> ignoredHeaders, Row row, String iri, ArrayList<Metadata> metadataArrayList) {

        // Transform array into hashmap
        HashMap<String, Metadata> metadataLookup = new HashMap<>();
        for (Metadata metadata : metadataArrayList) {
            metadataLookup.put(metadata.getLabel(), metadata);
        }

        Iterator<String> headerIter = header.iterator();
        // Parsing all the headers
        while (headerIter.hasNext()) {
            String headerItem = headerIter.next();
            if (headerItem == null) continue;
            // Skip processed headers
            if (!ignoredHeaders.contains(headerItem)) {
                // Predicate element
                String predicate = null;

                // Obtain the cell and if empty / null skip it
                Cell cell = row.getCell(header.indexOf(headerItem));
                if (cell == null) continue;

                // If the metadata key in the lookup list process accordingly otherwise process dynamically
                Metadata metadata = null;
                if (metadataLookup.containsKey(headerItem)) {
                    // logger.info("Found in metadata lookup hash " + headerItem );
                    metadata = metadataLookup.get(headerItem);
                    Pattern pattern = metadata.getRegex();
                    String value = getStringValueCell(cell);
                    // Clear out multi line strings
                    value = value.replaceAll("[\n\r]", " ");
                    // ignoring empty cells
                    if (metadata.getRequirement().toLowerCase().matches("m") && value.length() == 0) {
                        logging("There is no value for the obligatory field " + metadata.getLabel() + " and should match pattern " + metadata.getSyntax() + " such as in example " + metadata.getExample());
                        throw new IllegalFormatException(metadata.getSyntax() + "\n" + value);
                    }

                    // Ignore empty cells that are not obligatory
                    if (value.length() == 0) continue;

                    Matcher m = pattern.matcher(value);
                    // It passes the regex validation
                    if (m.matches()) {
                        predicate = metadata.getURL();
                    } else {
                        // Message for normal field and obligatory field
                        String additionalLine = " does not match the pattern of " + metadata.getSyntax() + " regex " + metadata.getRegex().pattern() + " such as in example \"" + metadata.getExample() + "\"";
                        String prefixLine = "The value \"" + value + "\" of \"" + metadata.getLabel();
                        if (metadata.getRequirement().toLowerCase().contains("m")) {
                            logging(prefixLine + "\" which is obligatory " + additionalLine);
                        } else {
                            logging(prefixLine + additionalLine);
                        }

                        // Minimum length
                        if (pattern.pattern().matches(".*\\{[0-9]+,}.*")) {
                            logging("A minimum length ("+pattern.pattern().replaceAll(".*?\\{([0-9]+),}.*", "$1")+") is set for " + metadata.getLabel());
                        }
                        // Exact length
                        if (pattern.pattern().matches(".*\\{[0-9]+}.*")) {
                            logging("An exact length ("+pattern.pattern().replaceAll(".*?\\{([0-9]+)}.*", "$1")+") is set for " + metadata.getLabel());
                        }
                        // Range length
                        if (pattern.pattern().matches(".*\\{[0-9]+,[0-9]+}.*")) {
                            logging("A range length ("+pattern.pattern().replaceAll(".*?\\{([0-9]+,[0-9]+)}.*", "$1")+") is set for " + metadata.getLabel());
                        }

                        throw new IllegalFormatException(metadata.getSyntax() + "\n" + value);
                    }
                }

                // If no predicate was available from the excel sheet
                if (predicate != null && predicate.trim().length() == 0)
                    predicate = null;

                if (predicate == null) {
                    // Parsing elements from the standardisation and also optionally added by the user not in the metadata excel sheet

                    // Processing the header name as it can contain non acceptable chars
                    String beforeHeaderItem = iriCorrector(headerItem);

                    // Replace non ascii characters by closed character
                    headerItem = Normalizer.normalize(beforeHeaderItem, Normalizer.Form.NFD);
                    headerItem = headerItem.replaceAll("[^\\x00-\\x7F]", "");

                    // It is null if the user created this column by itself
                    if (predicate == null) {
                        logger.debug("No predicate defined for " + headerItem);
                        predicate = PREFIX + headerItem;
                    }
                }

                logger.debug("Predicate value: " + predicate);

                // If cell is not null / empty see if we can map to appropriate triple type
                switch (cell.getCellType()) {
                    case BOOLEAN:
                        domain.getRDFSimpleCon().add(iri, predicate, cell.getBooleanCellValue());
                        break;
                    case NUMERIC:
                        if (HSSFDateUtil.isCellDateFormatted(cell)) {
                            Date date = cell.getDateCellValue();
                            LocalDate localDate = date.toInstant().atZone(ZoneId.systemDefault()).toLocalDate();
                            domain.getRDFSimpleCon().add(iri, predicate, localDate);
                        } else {
                            if (cell.getNumericCellValue() % 1 == 0) {
                                domain.getRDFSimpleCon().add(iri, predicate, (int) cell.getNumericCellValue());
                            } else {
                                domain.getRDFSimpleCon().add(iri, predicate, cell.getNumericCellValue());
                            }
                        }
                        break;
                    case BLANK:
                    case ERROR:
                    case _NONE:
                    case FORMULA:
                        break;
                    case STRING:
                        if (getStringValueCell(cell).trim().trim().length() == 0) break;
                        domain.getRDFSimpleCon().addLit(iri, predicate, getStringValueCell(cell).trim().trim());
                        break;
                    default:
                        logging("Unknown cell type, please contact admin to improve the code : " + cell.getCellType().name());
                        domain.getRDFSimpleCon().addLit(iri, predicate, getStringValueCell(cell).trim().trim());
                }

                // Check if predicate already has a type property and
                // if not create the predicate with rdf type and label
                NodeIterator nodeIterator = domain.getRDFSimpleCon().getObjects(predicate, "rdf:type");
                if (!nodeIterator.hasNext()) {
                    // Add predicate RDFS statement
                    domain.getRDFSimpleCon().add(predicate, "rdf:type", "rdf:Property");
                    // TODO use label from new version of excel sheet?
                    if (metadata != null) {
                        domain.getRDFSimpleCon().addLit(predicate, "rdfs:label", metadata.getLabel());
                    } else {
                        String value = StringUtils.join(splitCamelCase(predicate.replaceAll(".*/", "")), " ").toLowerCase();
                        // System.err.println(" >>> " + value);
                        domain.getRDFSimpleCon().addLit(predicate, "rdfs:label", value);
                    }
                }
            }
        }
    }


    /**
     *
     * @param row Creates a column index lookup based on naming scheme
     * @return
     */
    private static ArrayList<String> getHeader(Row row) {
        ArrayList<String> headerLookup = new ArrayList<>();
        for (int j = 0; j < row.getLastCellNum(); j++) {
            Cell cell = row.getCell(j);
            try {
                headerLookup.add(cell.getStringCellValue().trim().trim());
            } catch (NullPointerException e) {
                headerLookup.add(null);
            }
        }
        return headerLookup;
    }

    public void setMessage(String message) {
        this.message = message;
    }

    public String getMessage() {
        return message;
    }
}
